import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.lines import Line2D
import matplotlib.animation as animation


class Phasor(object):
    """Phasor that can be advanced in time with rotation and growth rates.

    Parameters
    ----------
    init : complex
        Initial phasor in rectangular form (Re + jIm)
    frequency : float, optional
        Rotation rate in rad/s.
    growth_rate : float, optional
        Exponential growth rate (decay if < 0).

    Attributes
    ----------
    t : float
        Current time.
    re : float
        Current real component of the phasor.
    im : float
        Current imaginary component of the phasor.
    radius : float
        Current radius of the phasor.
    angle : float
        Current angle of the phasor.
    trace_t : list
        History of time values (since most recent `clear()`).
    trace_re : list
        History of real component values (since most recent `clear()`).
    trace_im : list
        History of imaginary component values (since most recent `clear()`).
    """

    def __init__(self, init, frequency=0, growth_rate=0):
        self.init = init
        self.frequency = frequency
        self.growth_rate = growth_rate

        self.trace_t = []
        self.trace_re = []
        self.trace_im = []

        self._init()

    @classmethod
    def from_eig(cls, eigvec_component, eigval):
        """Creates a phasor from an eigenvalue/eigenvector component pair.

        Parameters
        ----------
        eigvec_component : complex
            A single eigenvector component representing the phasor's initial
            real/imaginary parts.
        eigval : complex
            The eigenvector, which specifies the phasor's growth rate (real
            part) and rotational frequency (imaginary part).
        """
        return cls(eigvec_component, np.imag(eigval), np.real(eigval))

    def advance(self, dt):
        """Advance the phasor by a time step dt."""
        self.t += dt
        self.radius *= np.exp(self.growth_rate * dt)
        self.angle += self.frequency * dt

        self._update_rect()

        self.trace_t.append(self.t)
        self.trace_re.append(self.re)
        self.trace_im.append(self.im)

    def clear(self):
        """Clear trajectories."""
        self._init()
        del self.trace_t[:]
        del self.trace_re[:]
        del self.trace_im[:]

    def _init(self):
        """Initialize parameters."""
        self.t = 0
        self.radius = np.abs(self.init)
        self.angle = np.angle(self.init)
        self._update_rect()

    def _update_rect(self):
        """Update rectangular components."""
        vec = self.radius * np.exp(1j * self.angle)
        self.re = np.real(vec)
        self.im = np.imag(vec)


class PhasorAnimation(animation.TimedAnimation):
    """Animation for demonstrating rotating phasors.

    Two axes are set up. On top, there is an s-plane to show the real and
    imaginary components of the phasors. The current phasor "vector" is shown
    with a thick line, the current endpoint of the vector is shown with a
    circle, thin lines show the projection of the real part of the phasor down
    to the bottom of the plane, and the time history of the endpoint of the
    vectors are shown.

    On bottom, the phasors' real components are plotted in time. The plot is
    rotated so that time is positive downward, and the x axes of the s-plane
    and the time plots are lined up. The current value is shown with a circle,
    thin lines show the projection from the top of the plot to the current
    value, and the time history is plotted.

    Parameters
    ----------
    fig : Figure
        matplotlib Figure object on which to animate.
    t : array
        Array of time values at which to plot. Even time spacing is assumed.
    phasors : list
        List of Phasor objects to advance and plot.
    re_range : tuple, optional
        Limits of the real axis.
    im_range : tuple, optional
        Limits of the imaginary axis.
    repeat : bool, optional
        Specifies whether or not to repeat the animation once it finishes.
    repeat_delay : float, optional
        Amount of time to wait before repeating the animation in milliseconds.
    time_stretch : float, optional
        Multiplicative factor of the plotting interval. Increasing
        `time_stretch` effectively makes the animation slower without affecting
        the time units.
    blit : bool, optional
        Specifies whether or not to use blitting.
    """

    def __init__(self, fig, t, phasors, re_range=(-1, 1), im_range=(-1, 1),
                 repeat=True, repeat_delay=0, time_stretch=1, blit=True):
        self.t = t
        self.dt = t[1] - t[0]
        self.phasors = phasors
        self.re_range = re_range
        self.im_range = im_range

        gs = GridSpec(2, 1, height_ratios=[1, 2])

        # s-plane plot of phasors
        ax_s = fig.add_subplot(gs[0])
        ax_s.set_ylabel('Im')
        ax_s.set_xlim(*re_range)
        ax_s.set_ylim(*im_range)
        ax_s.set_xticklabels(ax_s.get_xticklabels(), visible=False)
        ax_s.grid()

        # time plot of the real part of the phasors
        ax_t = fig.add_subplot(gs[1])
        ax_t.set_xlabel('Re')
        ax_t.set_ylabel('t')
        ax_t.set_xlim(*re_range)
        ax_t.set_ylim(self.t[0], self.t[-1]+self.dt)
        ax_t.invert_yaxis()
        ax_t.grid()

        fig.subplots_adjust(hspace=0)
        fig.tight_layout()

        # vectors in the Re/Im axis from origin
        self.vec_lines = []
        # dots at the end of the vectors in the Re/Im axis
        self.vec_dots = []
        # lines streaming from the endpoints of the vectors to the axis base
        self.vec_connector_lines = []
        # trace of vectors in Re/Im axis
        self.vec_trace_lines = []
        # dot showing current x(t) value
        self.time_dots = []
        # lines streaming from the current x(t) value to the top of the axis
        self.time_connector_lines = []
        # trace of x(t) values
        self.time_trace_lines = []

        color_vals = np.linspace(0, 1, 10)
        for i, phasor in zip(color_vals, phasors):
            c = plt.cm.Set1(i)

            vl = Line2D([], [], color=c, linewidth=2)
            self.vec_lines.append(vl)
            ax_s.add_line(vl)

            vcl = Line2D([], [], color=c, linewidth=0.5)
            self.vec_connector_lines.append(vcl)
            ax_s.add_line(vcl)

            vd = Line2D([], [], color=c, marker='o', linewidth=0)
            self.vec_dots.append(vd)
            ax_s.add_line(vd)

            vtl = Line2D([], [], color=c)
            self.vec_trace_lines.append(vtl)
            ax_s.add_line(vtl)

            td = Line2D([], [], color=c, marker='o', linewidth=0)
            self.time_dots.append(td)
            ax_t.add_line(td)

            tcl = Line2D([], [], color=c, linewidth=0.5)
            self.time_connector_lines.append(tcl)
            ax_t.add_line(tcl)

            ttl = Line2D([], [], color=c)
            self.time_trace_lines.append(ttl)
            ax_t.add_line(ttl)

        animation.TimedAnimation.__init__(
            self, fig, interval=time_stretch/self.dt, blit=blit,
            repeat=repeat, repeat_delay=repeat_delay)

    def new_frame_seq(self):
        return iter(range(self.t.size))

    def _draw_frame(self, framedata):
        self._drawn_artists = []

        # advance phasors and plot them
        for i, phasor in enumerate(self.phasors):
            phasor.advance(self.dt)
            self.vec_lines[i].set_data([0, phasor.re], [0, phasor.im])
            self.vec_dots[i].set_data(phasor.re, phasor.im)
            self.vec_connector_lines[i].set_data([phasor.re, phasor.re],
                                                 [phasor.im, self.im_range[0]])
            self.vec_trace_lines[i].set_data(phasor.trace_re, phasor.trace_im)
            self.time_dots[i].set_data(phasor.re, phasor.t)
            self.time_connector_lines[i].set_data([phasor.re, phasor.re],
                                                  [0, phasor.t])
            self.time_trace_lines[i].set_data(phasor.trace_re, phasor.trace_t)

        # add lines to _drawn_artists
        for phasor in self.phasors:
            self._drawn_artists.extend(self.vec_lines)
            self._drawn_artists.extend(self.vec_dots)
            self._drawn_artists.extend(self.vec_connector_lines)
            self._drawn_artists.extend(self.vec_trace_lines)
            self._drawn_artists.extend(self.time_dots)
            self._drawn_artists.extend(self.time_connector_lines)
            self._drawn_artists.extend(self.time_trace_lines)

    def _init_draw(self):
        # clear the phasor trajectories
        for phasor in self.phasors:
            phasor.clear()

        # reset line data
        if getattr(self, '_drawn_artists', None) is not None:
            for a in self._drawn_artists:
                a.set_data([], [])


if __name__ == '__main__':
    times = np.arange(0, 3, 0.03)

    a = -0.2
    w = np.pi

    phasors = [
        Phasor(0.5+0.1j, frequency=w, growth_rate=a),
        Phasor(-0.5-0.1j, frequency=w, growth_rate=a),
        Phasor(0.2+0.0j, frequency=w, growth_rate=a),
        Phasor.from_eig(0.2-0.4j, a+1j*w)
    ]

    fig = plt.figure()
    ani = PhasorAnimation(fig, times, phasors)
    plt.show()
